#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/***   GENERAL PROGRAM SETTINGS   ***/

int window_size;
int num_features;
char* output_file;
char* flag_file;
char* thread_name;

int num_test_pro;
int max_test_len;
char* input_test;

int num_train_pro;
int max_train_len;
char* input_train;

/***   PROTEIN PROFILES & WINDOWS   ***/

struct Protein
{
  int len;
  double **pro;
  int win;
  double *nor;
  double *sim;
};

struct Protein **TEST;
double **product;

/***   TEST DATASET LOADING   ***/

void load_test_dataset_into_memory()
{
  TEST = malloc( num_test_pro * sizeof(struct Protein *) );
  FILE *infile = fopen(input_test,"r"); int protein_len;
  int num_windows; int protein; int pos; int feat;
  for(protein=0;protein<num_test_pro;protein++){
    TEST[protein] = malloc( sizeof(struct Protein) );
    fscanf(infile,"%d\n",&protein_len);
    fscanf(infile,"%d\n",&num_windows);
    TEST[protein]->len = protein_len;
    TEST[protein]->win = num_windows;
    TEST[protein]->pro = malloc( protein_len * sizeof(double *) );
    for(pos=0;pos<protein_len;pos++){
      TEST[protein]->pro[pos] = malloc( num_features * sizeof(double) );
      for(feat=0;feat<num_features;feat++){
        fscanf(infile,"%lf\n",&TEST[protein]->pro[pos][feat]); } } }
  fclose(infile);
}

/***   INITIALIZATION & ALLOCATION   ***/

void initialize_windows_and_allocate()
{
  int protein; int protein_len; int num_windows; int window;
  int pos; int feat; double sum; double contrib[max_test_len];
  for(protein=0;protein<num_test_pro;protein++){
    protein_len = TEST[protein]->len;
    for(pos=0;pos<protein_len;pos++){ sum = 0;
      for(feat=0;feat<num_features;feat++){
        sum+=TEST[protein]->pro[pos][feat]*TEST[protein]->pro[pos][feat]; }
      contrib[pos] = sum; } num_windows = TEST[protein]->win;
    TEST[protein]->nor = malloc( num_windows * sizeof(double) );
    TEST[protein]->sim = malloc( num_windows * sizeof(double) );
    for(window=0;window<num_windows;window++){
      TEST[protein]->sim[window] = -1.0; sum = 0;
      for(pos=0;pos<window_size;pos++){
        sum+=contrib[window+pos]; }
      TEST[protein]->nor[window] = sqrt(sum); } }
  product = malloc( max_test_len * sizeof(double *) );
  for(pos=0;pos<max_test_len;pos++){
    product[pos] = malloc( max_train_len * sizeof(double) ); }
}

/***   COSINE SIMILARITY ANALYSIS   ***/

void run_cosine_similarity_analysis()
{
  
  FILE *infile = fopen(input_train,"r");
  int feat; int i; double sum; double sim;
  
  int train_index;
  int train_len;
  int train_pos;
  int train_win;
  int train_beg;
  double train_pro[max_train_len][num_features];
  double train_contrib[max_train_len];
  double train_nor[max_train_len];
  
  int test_index;
  int test_len;
  int test_pos;
  int test_win;
  int test_beg;
  double test_win_nor;
  double test_win_sim;
  double **test_pro;
  double *test_nor;
  double *test_sim;
  
  /***   LOAD TRAINING PROFILE   ***/
  
  for(train_index=0;train_index<num_train_pro;train_index++)
  {
    fscanf(infile,"%d\n",&train_len);
    fscanf(infile,"%d\n",&train_win);
    for(train_pos=0;train_pos<train_len;train_pos++){
      sum=0; for(feat=0;feat<num_features;feat++){
        fscanf(infile,"%lf\n",&train_pro[train_pos][feat]);
        sum+=train_pro[train_pos][feat]*train_pro[train_pos][feat]; }
      train_contrib[train_pos]=sum; }
    for(train_beg=0;train_beg<train_win;train_beg++){
      sum=0; for(train_pos=0;train_pos<window_size;train_pos++){
        sum+=train_contrib[train_beg+train_pos]; }
      train_nor[train_beg] = sqrt(sum); }
    
    /***   COMPUTE PARTIAL DOT PRODUCTS   ***/
    
    fprintf(stderr,"    [%s] Processing training protein %d / %d (len = %d)\n",
      thread_name,train_index+1,num_train_pro,train_len);
    for(test_index=0;test_index<num_test_pro;test_index++)
    {
      test_len = TEST[test_index]->len;
      test_win = TEST[test_index]->win;
      test_pro = TEST[test_index]->pro;
      test_nor = TEST[test_index]->nor;
      test_sim = TEST[test_index]->sim;
      for(test_pos=0;test_pos<test_len;test_pos++){
        for(train_pos=0;train_pos<train_len;train_pos++){
          sum=0; for(feat=0;feat<num_features;feat++){
            sum+=test_pro[test_pos][feat]*train_pro[train_pos][feat]; }
          product[test_pos][train_pos] = sum; } }
      
      /***   UPDATE SIMILARITY RESULTS   ***/
      
      for(test_beg=0;test_beg<test_win;test_beg++){
        test_win_nor = test_nor[test_beg];
        test_win_sim = test_sim[test_beg];
        for(train_beg=0;train_beg<train_win;train_beg++){
          sum=0; for(i=0;i<window_size;i++){
            sum+=product[test_beg+i][train_beg+i]; }
          sim=sum/(test_win_nor*train_nor[train_beg]);
          if(sim > test_win_sim){
            test_win_sim = sim; } }
        test_sim[test_beg] = test_win_sim; }
    }
  }
  fclose(infile);
}

/***   WRITE OUTPUT FILE   ***/

void write_cosine_similarity_results()
{
  
  FILE *outfile = fopen(output_file,"w");
  int protein;
  int num_window;
  int window;
  
  for(protein=0;protein<num_test_pro;protein++){
    num_window = TEST[protein]->win;
    for(window=0;window<num_window;window++){
      fprintf(outfile,"%.12f\n",TEST[protein]->sim[window]); } }
  fclose(outfile);
  
  FILE *ffile = fopen(flag_file,"w");
  fprintf(ffile,"done\n");
  fclose(outfile);
  
}

/***   PROGRAM MAIN FUNCTION   ***/

int main(int argc,char* argv[])
{
  thread_name = argv[1];
  window_size = atoi(argv[2]);
  num_features = atoi(argv[3]);
  output_file = argv[4];
  flag_file = argv[5];
  num_test_pro = atoi(argv[6]);
  max_test_len = atoi(argv[7]);
  input_test = argv[8];
  num_train_pro = atoi(argv[9]);
  max_train_len = atoi(argv[10]);
  input_train = argv[11];
  load_test_dataset_into_memory();
  initialize_windows_and_allocate();
  run_cosine_similarity_analysis();
  write_cosine_similarity_results();
  return 0;
}

