#!/usr/bin/perl
use strict;

####################################################################################################

my $pairwise_exe="./multicore_analysis"; my $script_usage="./multicore_analysis.pl";
$script_usage.=" training_data test_data output_file window_size num_threads";
if($#ARGV!=4){ print "\n  PROGRAM USAGE :\n\n  $script_usage\n\n"; exit 1; }
my $input_train=$ARGV[0]; my $input_test=$ARGV[1]; my $output_file=$ARGV[2];
my $window_size=$ARGV[3]; my $num_threads=$ARGV[4];
if(! -f $pairwise_exe){ print "\n  C program not found, aborting...\n\n"; exit 1; }
if(! -f $input_train){ print "\n  Training data not found, aborting...\n\n"; exit 1; }
if(! -f $input_test){ print "\n  Test data not found, aborting...\n\n"; exit 1; }
open(OUT,">$output_file") || die "\n  Cannot write output file, aborting...\n\n";
close(OUT); `rm $output_file`; if($window_size!~/^[1-9][0-9]$/){
print "\n  Window size must be at least 10 & at most 99\n\n"; exit 1; }
my $pairwise_test="$output_file.test"; if($num_threads!~/^[1-9][0-9]?$/){
print "\n  Num threads must be at least 1 & at most 99\n\n"; exit 1; }
print "\n  Program arguments successfully checked, proceeding...\n";

####################################################################################################

my $num_features=0; if($num_features==0){ open(IN,"$input_train"); my $header=<IN>;
if($header!~/^>/){ print "\n  Inconsistent input file format, aborting...\n\n";
exit 1; } my $line=<IN>; chomp($line); my @FEATURES=split(" ",$line); close(IN);
while($FEATURES[$#FEATURES] eq ""){ pop(@FEATURES); } $num_features=$#FEATURES+1; }
if(($num_features!~/^[1-9][0-9]*$/)||($num_features<3)||($num_features>999)){
print "\n  Number of features must be at least 3 & at most 999\n\n"; exit 1; }
print "  Detected number of features / protein position: $num_features...\n";

####################################################################################################

my $num_BATCH=0;   my @BATCH_trah;   my $num_train_pro=0;   my $num_PROTEIN=0;
my @BATCH_name;    my @BATCH_test;   my $num_train_win=0;   my @PROTEIN_id;
my @BATCH_numw;    my @BATCH_outf;   my $max_train_len=0;   my @PROTEIN_len;
my @BATCH_nump;    my @BATCH_outh;   my $num_test_pro=0;    my @PROTEIN_win;
my @BATCH_maxl;    my @BATCH_flag;   my $num_test_win=0;    my @PROTEIN_acc;
my @BATCH_traf;    my @BATCH_done;   my $max_test_len=0;

####################################################################################################

print "  Preprocessing test set for heavy part of analysis...\n"; open(IN,"$input_test");
open(OUT,">$pairwise_test"); my $header=<IN>; chomp($header); while(my $l=<IN>){
chomp($l); my $protein_id=substr($header,1); my @ACC; my @PRO; my @d=split(" ",$l);
if($header!~/^>/){ print "\n  Inconsistent input file format, aborting...\n\n"; exit 1; }
if($#d!=($num_features+1)){ print "\n  Inconsistent input file format, aborting...\n\n";
exit 1;}if($d[0] eq $d[1]){$ACC[0]=1;} else{$ACC[0]=0;} for(my $i=0;$i<$num_features;$i++){
$PRO[0][$i]=$d[$i+2]; } my $protein_len=1; my $done=0; while($done==0){ if($l=<IN>){ chomp($l);
if($l=~/^>/){ $header=$l; $done=1; } else{ my @T=split(" ",$l); if($#T!=($num_features+1)){
print "\n  Inconsistent input file format, aborting...\n\n"; exit 1; } if($T[0] eq $T[1]){
$ACC[$protein_len]=1; } else{ $ACC[$protein_len]=0; } for(my $i=0;$i<$num_features;$i++){
$PRO[$protein_len][$i]=$T[$i+2]; } $protein_len++; } } else{ $done=1; } }
if($protein_len>=$window_size){ my $windows=$protein_len-$window_size+1;
$PROTEIN_id[$num_PROTEIN]=$protein_id; $PROTEIN_len[$num_PROTEIN]=$protein_len;
$PROTEIN_win[$num_PROTEIN]=$windows; $PROTEIN_acc[$num_PROTEIN]=join("",@ACC);
$num_PROTEIN++; $num_test_pro++; $num_test_win+=$windows; if($protein_len>$max_test_len){
$max_test_len=$protein_len; } print OUT "$protein_len\n"; print OUT "$windows\n";
for(my $pos=0;$pos<$protein_len;$pos++){ for(my $i=0;$i<$num_features;$i++){
print OUT "$PRO[$pos][$i]\n"; } } } else{ print "    [WARNING] Protein $protein_id ";
print "ignored (shorter than requested window size)\n"; } } close(OUT); close(IN);
print "  Dataset found with $num_test_pro proteins & $num_test_win windows...\n";

####################################################################################################

print "  Preprocessing train set for heavy part of analysis...\n";
for(my $batch_index=0;$batch_index<$num_threads;$batch_index++){ $num_BATCH++;
$BATCH_name[$batch_index]="thread_$batch_index"; $BATCH_numw[$batch_index]=0;
$BATCH_nump[$batch_index]=0; $BATCH_traf[$batch_index]="$output_file.train.$batch_index";
$BATCH_maxl[$batch_index]=0; local *FILE; open(FILE,">$BATCH_traf[$batch_index]");
$BATCH_trah[$batch_index]=*FILE; $BATCH_test[$batch_index]="$pairwise_test";
$BATCH_outf[$batch_index]="$output_file.out.$batch_index"; $BATCH_done[$batch_index]=0;
$BATCH_flag[$batch_index]="$output_file.done.$batch_index"; } open(IN,"$input_train");
$header=<IN>; chomp($header); while(my $l=<IN>){ chomp($l); if($header!~/^>/){
print "\n  Inconsistent input file format, aborting...\n\n"; exit 1; }
my $protein_id=substr($header,1);my @PRO;my @d=split(" ",$l);if(($#d+1)!=$num_features){
print "\n  Inconsistent input file format, aborting...\n\n"; exit 1; } my $protein_len=1;
for(my $i=0;$i<$num_features;$i++){ $PRO[0][$i]=$d[$i]; } my $done=0; while($done==0){
if($l=<IN>){ chomp($l); if($l=~/^>/){ $header=$l; $done=1; } else{ my @T=split(" ",$l);
if(($#T+1)!=$num_features){ print "\n  Inconsistent file format, aborting...\n\n"; exit 1; }
for(my $i=0;$i<$num_features;$i++){ $PRO[$protein_len][$i]=$T[$i]; } $protein_len++; } }

####################################################################################################

else{ $done=1; } } if($protein_len>=$window_size){ my $windows=$protein_len-$window_size+1;
my $batch_index=0;for(my $i=1;$i<$num_threads;$i++){if($BATCH_numw[$i]<$BATCH_numw[$batch_index]){
$batch_index=$i; } } $BATCH_numw[$batch_index]+=$windows; $BATCH_nump[$batch_index]++;
if($protein_len>$BATCH_maxl[$batch_index]){ $BATCH_maxl[$batch_index]=$protein_len; }
my $batch_trah=$BATCH_trah[$batch_index]; $num_train_pro++; $num_train_win+=$windows;
if($protein_len>$max_train_len){ $max_train_len=$protein_len; }
print $batch_trah "$protein_len\n$windows\n"; for(my $pos=0;$pos<$protein_len;$pos++){
for(my $i=0;$i<$num_features;$i++){ print $batch_trah "$PRO[$pos][$i]\n"; } } } else{
print "    [WARNING] Protein $protein_id ignored (shorter than requested window size)\n"; } }
close(IN); print "  Dataset found with $num_train_pro proteins & $num_train_win windows...\n";
if($num_test_pro==0){ print "\n  Empty testing set, cannot proceed...\n\n"; exit 1; }
if($num_train_pro==0){ print "\n  Empty training set, cannot proceed...\n\n"; exit 1; }
if($num_train_pro<$num_threads){ print "\n  Too many threads requested...\n\n"; exit 1; }
for(my $batch_ind=0;$batch_ind<$num_threads;$batch_ind++){ close($BATCH_trah[$batch_ind]); }

####################################################################################################

print "  Now running cosine similarity analysis heavy part...\n\n";
for(my $batch_ind=0;$batch_ind<$num_threads;$batch_ind++){
my $basic_args="$BATCH_name[$batch_ind] $window_size $num_features";
my $output_args="$BATCH_outf[$batch_ind] $BATCH_flag[$batch_ind]";
my $test_args="$num_test_pro $max_test_len $BATCH_test[$batch_ind]";
my $train_args="$BATCH_nump[$batch_ind] $BATCH_maxl[$batch_ind] $BATCH_traf[$batch_ind]";
my $command_line="$pairwise_exe $basic_args $output_args $test_args $train_args";
system("$command_line &"); } my $jobs_completed=0; while($jobs_completed==0){
my $num_completed=0; for(my $batch_ind=0;$batch_ind<$num_threads;$batch_ind++){
if(-f $BATCH_flag[$batch_ind]){ $num_completed++; } }
if($num_completed<$num_threads){ sleep(5); } else{ $jobs_completed=1; } }
print "\n  Done running cosine similarity analysis heavy part...\n";
print "  Checking analysis results and writing output file...\n\n";

####################################################################################################

for(my $batch_ind=0;$batch_ind<$num_threads;$batch_ind++){ if(! -f "$BATCH_outf[$batch_ind]"){
print "\n  Cannot find analysis results from C program, exiting...\n\n"; exit 1; }
local *FILE; open(FILE,"$BATCH_outf[$batch_ind]"); $BATCH_outh[$batch_ind]=*FILE; }
open(OUT,">$output_file"); print OUT "ProteinID\tWindowStartPos\tCosineSimilarity\tAccuracy\n";
for(my $protein_ind=0;$protein_ind<$num_PROTEIN;$protein_ind++){
my $num_windows=$PROTEIN_win[$protein_ind]; my @ACC=split("",$PROTEIN_acc[$protein_ind]);
my $protein_id=$PROTEIN_id[$protein_ind];for(my $win_ind=0;$win_ind<$num_windows;$win_ind++){
my $accuracy=0; my $cos_sim=-1; for(my $batch=0;$batch<$num_threads;$batch++){
my $value=readline($BATCH_outh[$batch]); chomp($value); if($value!~/^-?[01]\.[0-9]{12}$/){
print "\n  Inconsistent cosine similarity value: $value\n\n"; exit 1; } if($value>$cos_sim){
$cos_sim=$value; } } my $rounded=sprintf("%.12f",$cos_sim); if(($rounded<-1)||($rounded>1)){
print "\n  Inconsistent similarity: $cos_sim\n\n"; exit 1; } for(my $i=0;$i<$window_size;$i++){
$accuracy+=$ACC[$win_ind+$i]; } $accuracy=sprintf("%.6f",$accuracy/$window_size);
print OUT "$protein_id\t".($win_ind+1)."\t$rounded\t$accuracy\n"; } } close(OUT);
for(my $batch=0;$batch<$num_threads;$batch++){ if(my $l=readline($BATCH_outh[$batch])){
print "\n  Inconsistent result file length, aborting\n\n"; exit 1; } close($BATCH_outh[$batch]);
`rm -f $BATCH_traf[$batch] $BATCH_outf[$batch] $BATCH_flag[$batch]`; } `rm -f $pairwise_test`;

####################################################################################################

